unit NCommand;

{$IMPORTEDDATA OFF}

interface

uses Windows, NDict;

type
  TCommandProc = procedure of object;

  TCommand = class
  private
    Commands: array of TCommandProc;
    function NameToId(const Name: string): Integer;
  public
    AccelHandle: HACCEL;
    constructor Create(const Commands: array of TCommandProc);
    destructor Destroy; override;
    function Id(const Proc: TCommandProc): Integer;
    procedure AddMenuItems(MenuHandle: HMENU; const Items: array of string);
    procedure BindKeys(const KeyMap: TStrArrayDict);
    procedure DispatchCommand(Id: Integer);
  end;

implementation

uses NKeyMap;

const
  FirstId = 20000;
  CommandProcPrefix = 'cm';

constructor TCommand.Create(const Commands: array of TCommandProc);
var
  i: Integer;
begin
  SetLength(Self.Commands, Length(Commands));
  for i := 0 to High(Self.Commands) do
    Self.Commands[i] := Commands[i];
end;

destructor TCommand.Destroy;
begin
  if AccelHandle <> 0 then
    DestroyAcceleratorTable(AccelHandle);
end;

procedure TCommand.AddMenuItems(MenuHandle: HMENU;
    const Items: array of string);
  function AccelToKeyName(const Accel: TAccel): string;
  var
    ks: TKeyState;
  begin
    ks := [];
    if (Accel.fVirt and FALT)     <> 0 then Include(ks, ksAlt);
    if (Accel.fVirt and FSHIFT)   <> 0 then Include(ks, ksShift);
    if (Accel.fVirt and FCONTROL) <> 0 then Include(ks, ksCtrl);
    Result := VkeyToKeyName(ks, Accel.key);
  end;
var
  i, j, id: Integer;
  capt, name, keyname: string;
  accels: array of TAccel;
  menu: HMENU;
begin
  menu := MenuHandle;
  SetLength(accels, CopyAcceleratorTableW(AccelHandle, Pointer(nil)^, 0));
  CopyAcceleratorTableW(AccelHandle, accels[0], Length(accels));

  i := 0;
  while i <= High(Items) do begin
    capt := Items[i];
    name := Items[i+1];
    if name = '' then
      menu := MenuHandle
    else if name = '-' then
      AppendMenuW(menu, MF_SEPARATOR, 0, nil)
    else if name[1] = '>' then begin
      menu := CreatePopupMenu;
      AppendMenuW(MenuHandle, MF_POPUP, menu, PWideChar(WideString(capt)));
    end else begin
      keyname := '';
      id := NameToId(name);
      for j := 0 to High(accels) do
        if accels[j].cmd = id then
          if keyname = '' then
            keyname := AccelToKeyName(accels[j])
          else
            keyname := keyname + ', ' + AccelToKeyName(accels[j]);
      AppendMenuW(menu, MF_STRING, id, PWideChar(WideString(capt + #9 + keyname)));
    end;
    Inc(i, 2);
  end;
end;

procedure TCommand.Bindkeys(const KeyMap: TStrArrayDict);
var
  accels: array of TAccel;
  i, id: Integer;
  key: string;
  ks: TKeyState;
  virt: Byte;
  vkey: Word;
begin
  SetLength(accels, KeyMap.Count);
  i := 0;
  while KeyMap.EachKeys(key) do begin
    if Length(KeyMap[key]) = 0 then
      Continue;
    id := NameToId(KeyMap[key][0]);
    if id = 0 then
      Continue;
    StrToVkey(key, ks, vkey);
    virt := FVIRTKEY;
    if ksAlt   in ks then virt := virt or FALT;
    if ksCtrl  in ks then virt := virt or FCONTROL;
    if ksShift in ks then virt := virt or FSHIFT;
    accels[i].key := vkey;
    accels[i].fVirt := virt;
    accels[i].cmd := id;
    Inc(i);
  end;
  if AccelHandle <> 0 then
    DestroyAcceleratorTable(AccelHandle);
  if i > 0 then
    AccelHandle := CreateAcceleratorTableW(accels[0], i)
  else
    AccelHandle := 0;
end;

function TCommand.NameToId(const Name: string): Integer;
var
  code: Pointer;
  i: Integer;
begin
  code := TObject(TMethod(Commands[0]).Data).MethodAddress(CommandProcPrefix + Name);
  for i := 0 to High(Commands) do
    if TMethod(Commands[i]).Code = code then begin
      Result := FirstId + i;
      Exit;
    end;
  Result := 0;
end;

function TCommand.Id(const Proc: TCommandProc): Integer;
var
  i: Integer;
begin
  for i := 0 to High(Commands) do
    if TMethod(Commands[i]).Code = TMethod(Proc).Code then begin
      Result := FirstId + i;
      Exit;
    end;
  Result := 0;
end;

procedure TCommand.DispatchCommand(Id: Integer);
begin
  Commands[Id - FirstId];
end;

end.
